; This file implements the IR timer code base for Sean Petzold.
; Copyright 2007 Peter K. Keller (psilord@cs.wisc.edu)

; Using the 4MHz internal oscillator means each instruction takes 1 microsecond,
; except for gotos and other branching instructions, which take two.

(in-package #:16f688)

(defconstant +spbrg_val+ 12)
;; (defconstant +spbrg_val+ 25)	; 9600 baud
;; (defconstant +spbrg_val+ 207)	; 1200 baud

; Bit values for beam_status
(defconstant BROKEN 0)		; is the beam currently broken, set of true
(defconstant PREVIOUS 1)	; the previous beam state, set if previously broken

; Bit values for irled_status
(defconstant ACTIVE 0)	 ; set if IRLED is on

(with-pic 

(config _CP_OFF _CPD_OFF _BOD_OFF _PWRTE_ON _WDT_OFF 
	_INTRC_OSC_NOCLKOUT _MCLRE_ON _FCMEN_OFF _IESO_OFF)

; User defined variables and their needed sizes
; Must *always* use $ with variable names in a cblock/udata/res, etc.
(cblock #x20
	dvar 1			; used for Delay Loop
	dvar2 1			; used for Delay Loop
	tmp0 1
	tmp1 1
	adc_hi 1		; storage for ADC high byte
	adc_lo 1		; storage for ADC low byte
	irled_status 1	; 0 for off, 1 for on
	beam_status 1	; various bit values for things about the beam
	calib_on 1		; MSB 8-bits of the calibration ADC value for irled on
	calib_off 1		; MSB 8-bits of the calibration ADC value for irled off
	calib_val 1)	; calib value which, when an adc sample is compared to it,
					; it denotes the beam is present or broken.

; Shared Bank 0 area in all banks.
(cblock #x70
	w_temp 1)		; Storage for ISR routine
	status_temp 1
	pclath_temp 1)

;**********************************************************************
	(.org		#x000)			; processor reset vector

	;; must always use @ with address labels...
	(.goto		@main)			; go to beginning of program

	(.org		#x004)			; interrupt vector location
	(.movwf		$w_temp)			; save off current W register contents
	(.ORG		#x005)			; part of the interrupt vector
	(.movf		STATUS W)		; move status register into W register
	(.movwf		$status_temp)	; save off contents of STATUS register
	(.movf		PCLATH W)		; move pclath register into W register
	(.movwf		$pclath_temp)	; save off contents of PCLATH register

; for now, just ignore any interrupts we get

	(.movf		$pclath_temp W)	; retrieve copy of PCLATH register
	(.movwf		PCLATH)			; restore pre-isr PCLATH register contents	
	(.movf		$status_temp W)	; retrieve copy of STATUS register
	(.movwf		STATUS)			; restore pre-isr STATUS register contents
	(.swapf		$w_temp F)
	(.swapf		$w_temp W)		; restore pre-isr W register contents
	(.retfie)				; return from interrupt

@main
	(.bcf		STATUS RP0)		; Bank 0
	(.call		@InitVariables)		; set up the state of various variables
	(.call		@InitIO)				; set up all of my digital/analog pins 
								; and init the ADC
	(.call		@CalibrateIRBeam)	; Try over and over to calibrate the beam, 
								; telling the PC what it is doing.
@Loop
	(.call		@DebounceBeamState)
	(.call		@IsBeamBroken)
	(.call		@NotifyBeamChange)
	(.call		@DisplayADCResult)
	(.goto		@Loop)				; infinite loop

; A simple delay function for ~1.4 milliseconds
@Delay_1ms
	(.movlw		#x01)				; set outer delay loop
	(.movwf		$dvar2)
@Delay0
	(.movlw		#xFF)
	(.movwf		$dvar)				; set inner delay loop
@Delay1
	(.decfsz	$dvar F)
	(.goto		@Delay1)

	(.decfsz	$dvar2 F)
	(.goto		@Delay0)
	
	(.return)

; Initialize various variables.
@InitVariables
	(.bcf		STATUS RP0)
	(.clrf		$beam_status)
	(.bcf		$beam_status BROKEN)	; the beam is not broken
	(.bcf		$beam_status PREVIOUS)	; the previous state is not broken
	(.return)

; Initialize the IO pins to be what I need them to be
@InitIO
	(.bcf		STATUS RP0)	; select bank 0
	(.movlw		#x07)		; turn off the comparators leaving the pins digital
	(.movwf		CMCON0)
	(.call		@InitPortA)
	(.call		@InitPortC)
	(.call		@InitSerial)
	(.call		@InitADC)
	(.bcf		STATUS RP0)	; back to bank 0
	(.return)

; For this function, we initialize PortA to have RA2/AN2 be configured as an
; analog input for ADC. RA3 is also a *mandatory* digital input pin.
@InitPortA
	(.bcf		STATUS RP0)	; Bank 0 for PORTA register
	(.clrf		PORTA)		; mark all lines as zero
	(.bsf		STATUS RP0)	; Bank 1 for ANSEL/TRISA registers
	(.movlw		#b00000100)	; ansel<2> is marked as analog input
	(.movwf		ANSEL
	(.movlw		#b00001100)	; ra3, ra2 are inputs. ra2 is now an analog input
	(.movwf		TRISA)
	(.return)

; For this function, all pins of PortC are configured as digital except rc5, which
; will be the serial rx port.
@InitPortC
	(.bcf		STATUS RP0)	; Bank 0 for PORTC reg
	(.clrf		PORTC)		; mark all lines as zero
	(.bsf		STATUS RP0)	; Bank 1 for TRISC reg
	(.movlw		#b00110000)	; rc5:4 tristate for serial port, rest are outputs
	(.movwf		TRISC)		
	(.bcf		STATUS RP0)	; Back to bank 0				
	(.return)


; This function sets up and turns on the ADC. It assumes the InitPortX functions
; have already set up the correct inputs as analog
@InitADC
	(.bsf		STATUS RP0)	; Bank 1
	(.movlw		#b0001000)	; set ADC Fosc/8 clock
	(.movwf		ADCON1)
	(.bcf		STATUS RP0)	; Bank 0
	(.movlw		#b0001001)	; left justified, Vdd, AN2, not in progress, enabled
	(.movwf		ADCON0)
	(.call		@ADCDelay)	; wait for the ADC to stabalize
	(.return)

; set up the serial port for communications...
@InitSerial
	(.bcf		STATUS RP0)	; Bank 0

	; set up the baud rate
	(.movlw		+spbrg_val+)		; the baud I'm going to use
	(.movwf		SPBRG)

	; set up the transmit & control register
	(.bcf		STATUS RP0)	; Bank 0
	(.movlw		#b00100110)	; don't care, 8-bit, xmit enable, async,
							; sync break xmit compl., high speed, TSR empty, 
							; 9th bit
	(.movwf		TXSTA)

	; set up the receive & control register
	(.bcf		STATUS RP0)	; Bank 0
	(.movlw		#b10010000)	; serial port enable, 8 bit, don't care, en recv
							; don't care, no frame err, no overrun err, 
							; no 9th bit
	(.movwf		RCSTA)
	(.call		@Delay_1ms)	; make sure it stabalizes
	(.movf		RCREG W)	; clean out the usart recv buffer
	(.bcf		RCSTA CREN)	; clear any errors if there were any
	(.bsf		RCSTA CREN)
	(.return)

; This function does a polling sample of the ADC, then leaves the values in 
; adc_hi and adc_lo.
@SampleADC
	(.bcf		STATUS RP0)	; Bank 0
	(.bsf		ADCON0 GO)	; Start the acquisition

	(.btfsc		ADCON0 GO)	; Poll for completion
	(.goto		(1- @pc)
							; Grab the result out
	(.bcf		STATUS RP0)	; Bank 0
	(.movf		ADRESH W)	; Grab the high part
	(.movwf		$adc_hi)
	(.bsf		STATUS RP0	; Bank 1
	(.movf		ADRESL W)	; Grab the low part
	(.movwf		$adc_lo)

	(.return)

; This function assumes that the ADC has just been sampled. It sets beam_status
; to zero if, according to the ADC the beam has not been broken, and one if
; according to the ADC, the beam has been broken.
@IsBeamBroken
	(.bcf		STATUS RP0)				; Bank 0
	(.bcf		$beam_status BROKEN)	; assume no it isn't
	(.movf		$adc_hi W)
	(.subwf		$calib_val W)			; Perform: calib_val - W
										; If W is eq/smaller than calib_val,
										; then beam is *not* broken.
										; C is one if W <= calib_val
	(.btfss		STATUS C)				; is C zero (so W > calib_val)
	(.bsf		$beam_status BROKEN)	; yes, set broken bit
	
	(.return)

; This function samples the ADC with a delay in the middle and only if both
; samples agree is the beam considered stable
@DebounceBeamState
	(.bcf		STATUS RP0)		; bank 0
	(.movf		$beam_status W)	; save previous beam status
	(.movwf		$tmp0)
	(.call		@SampleADC)		; get the first reading
	(.call		@IsBeamBroken)	; check it
	(.movf		$beam_status W)
	(.andlw		(ash 1 BROKEN))	; look only at the broken/present bit
	(.movwf		$tmp1)			; save the beam state
	(.call		@Delay_1ms)		; wait ~1 millisecond
	(.call		@SampleADC)		; get the second reading
	(.call		@IsBeamBroken)	; check it
	(.movf		$beam_status W)
	(.andlw		(ash 1 BROKEN))	; look only at the broken/present bit
	(.subwf		$tmp1 W)		; check beam stability
	(.btfsc		STATUS Z)		; Is Z one (for the borrow)?
								; 0 - 0 = 0; 1 - 1 = 0
	(.goto		@DBS_debounced)	; yup, state is debounced
	(.movf		$tmp0 W)		; nope, so restore original beam state
	(.movwf		$beam_status)
	(.return)
@DBS_debounced
	(.movf		$tmp0 W)			; grab old beam state
	(.movwf		$beam_status)
	(.andlw		(ash 1 PREVIOUS))	; keep whatever the previous bit was
	(.iorwf		$tmp1 W)			; mix it with new beam status
	(.movwf		$beam_status)		; fix it up nice for IsBeamBroken
	(.return)

; Wait a minimum of 6 microseconds for the ADC to stabilize
@ADCDelay
	(.nop)
	(.nop)
	(.nop)
	(.nop)
	(.nop)
	(.nop)
	(.return)

; This function performs some averaging for an ADC sample when the beam is off
; and when the beam is on. This will give us voltage brackets which dictate
; the state of the beam given ambient IR conditions. If the beam cannot be
; calibrated, info concerning such will be output to the serial port and the
; calibration retried. Notice that when the irled is off, the ADC will read a
; high value due to the circuit design. When the irled is on, a low adc 
; value will be read.
@CalibrateIRBeam
	(.bcf		STATUS RP0)	; Bank 0
	(.movlw		#b00000000	; turn irled off (RA4)
	(.movwf		PORTA)
	(.call		@Delay_1ms)	; Wait for transistor & photodiode to stabilize
	(.call		@SampleADC)	; grab a sample
	(.bcf		STATUS RP0)	; Bank 0
	(.movf		$adc_hi W)
	(.movwf		$calib_off)	; store it in the irled off calibration value, the
							; voltage, and therefore the bit pattern is high 
							; in the case of an off irled
	
	(.movlw		#b00010000)	; turn beam on (RA4)
	(.movwf		PORTA)
	(.call		@Delay_1ms)	; wait for the transistor & photodiode to stabilize
	(.call		@SampleADC)	; grab a sample
	(.bcf		STATUS RP0)	; Bank 0
	(.movf		$adc_hi W)
	(.movwf		calib_on)	; store it in the irled on calibration value, the
							; voltage, and therefore the bit pattern is low 
							; in the case of an on irled

	; make sure the irled is left turned on for the rest of the execution
	(.clrf		$irled_status)
	(.incf		$irled_status F)

	; now perform the average of the two to get a "zero" that when the 
	; ADC sample crosses, it means the beam is on or off.
	; This next piece of code implements an average as:
	; calib_on/2 + calib_off/2 = calib_val

	; calib_on/2
	(.movf		$calib_on W)
	(.movwf		$tmp0)
	(.bcf		STATUS C)
	(.rrf		$tmp0 F)
	
	; calib_off/2
	(.movf		$calib_off W)
	(.movwf		$tmp1)
	(.bcf		STATUS C)
	(.rrf		$tmp1 F)

	; Add them (this will fit into an 8 bit value)
	(.bcf		STATUS C)	; Get rid of polluted carry bit
	(.movf		$tmp0 W)
	(.addwf		$tmp1 W)		; complete the average
	
	; store the calibrated value for later use
	(.movwf		$calib_val)

	; check the validity of the calibration data and retry if bad.
	; This is done by subtracting the low from the high and seeing
	; if there is enough "room" to have a good reading. Room is 
	; defined as the result of the subtraction being more than 25% 
	; of the total range. This implies a %12.5 range in the "on/off"
	; region around the average, which is only decimal number 13, a tight
	; but meaningful range. In this case it would be 256 * .25, which is
	; ~64. So, if the subtraction results in a number less than 64, we
	; know the calibration probably doesn't make sense
	(.movf		$calib_on W)
	(.subwf		$calib_off W)
	(.btfss		STATUS C)	; calib_on(low voltage) > calib_off(high voltage)?
	(.goto		@CIRB_retry)	; yes, retry
	(.movwf		$tmp0)			; nope, so do range check
	; check it against 64
	(.movlw		64)
	(.subwf		$tmp0 W)
	(.btfss		STATUS C)		; did the subtraction result in a borrow?
	(.goto		@CIRB_retry)	; yes, so retry
	(.call		@SendGoodCalib)	; nope, there is enough room so use the 
								; calibration
	(.call		@SendCalibData)
	(.return)

@CIRB_retry	; we lose, tell the pc what happened and try again.
	(.call		@SendBadCalib)
	(.call		@SendCalibData)
	(.call		@Delay_1ms)		; let things change a bit before retrying
	(.goto		@main)			; retry

	(.return)




; This function displays the ADC highest 4-bits, the pin format is:
; MSB RC3 RC2 RC1 RC0 LSB
; RA5 is beam broken/unbroken light
; RA4 is IR LED on/off
@DisplayADCResult
	(.bcf		STATUS RP0)				; Bank 0

	(.clrw)

	;; now, we ior in the beam broken/present bit
	(.iorlw		#b00100000)				; assume beam is unbroken, turn on ra5
	(.btfsc		$beam_status BROKEN)	; is the beam broken?
	(.andlw		#b11011111)				; yes, then turn off ra5


	;; Or in whether or not the IR emittor is turned on or not
	;; since the pin which controls the IR LED is also on this port.
	(.btfsc	irled_status ACTIVE)
	(.iorlw		#b00010000)	; turn on RA4 if irled is turned on

	(.movwf		PORTA)		; Shove it out to PORTA<1:0>

	;; Display the most significant 4 bits
	(.swapf		$adc_hi W)	; shove the high nibble into the low nibble spot
	(.andlw		#x0F)		; Only want the LSB 4 bits
	(.movwf		PORTC)		; Shove it out to PORTC<3:0>

	(.return)


;; If the calibration succeeded, then this must appear before the calibration 
;; data is written.

@SendGoodCalib
	(.movlw		#x43)				; C
	(.call		@TransmitSerial)	; Send the fact I calibrated successfully
	(.return)

; If, for whatever reason, the calibration, failed, this tells the PC about it.
@SendBadCalib
	(.movlw		#x4E)				; N
	(.call		@TransmitSerial)	; Send that I did not calibrate successfully
	(.return)

; Send a packet of calibration data (good or bad) to the PC
@SendCalibData
	(.movlw		#x4F)				; O for "on"
	(.call		@TransmitSerial)
	(.bcf		STATUS RP0)			; bank 0
	(.movf		$calib_on W)		; The on calibration byte
	(.call		@TransmitSerial)
	(.movlw		#x46)				; F for "off"
	(.call		@TransmitSerial)
	(.bcf		STATUS RP0)			; bank 0
	(.movf		$calib_off W)		; The off calibration byte
	(.call		@TransmitSerial)
	(.movlw		#x41)				; A for "average"
	(.call		@TransmitSerial)
	(.bcf		STATUS RP0)			; bank 0
	(.movf		$calib_val W)		; The calibration value
	(.call		@TransmitSerial)
	(.return)

@SendBeamResumeEvent
	(.movlw		#x52)				; R for beam restore event
	(.call		@TransmitSerial)
	(.return)

@SendBeamBreakEvent
	(.movlw		#x42)				; B for beam break event
	(.call		@TransmitSerial)
	(.return)


@NotifyBeamChange
	(.btfsc		$beam_status BROKEN)	; is beam broken?
	(.goto		@NBC_is_prev_broken)	; yes, check if previous was broken
	(.goto		@NBC_is_prev_resume)	; nope, check if previous was present

@NBC_is_prev_broken
	(.btfsc		$beam_status PREVIOUS)	; was the beam previously broken?
	(.return)							; yes, no event happened
	(.goto		@NBC_beam_break_event)	; nope, so a break event happened

@NBC_is_prev_resume
	(.btfsc		$beam_status PREVIOUS)	; was the beam previously broken?
	(.goto		@NBC_beam_resume_event)	; yes, so a resume event happend
	(.return)						; no, so nothing happened

@NBC_beam_break_event
	(.call		@SendBeamBreakEvent)
	(.bsf		$beam_status PREVIOUS)	; make previous match current
	(.return)

@NBC_beam_resume_event
	(.call		@SendBeamResumeEvent)
	(.bcf		$beam_status PREVIOUS)	; make previous match current
	(.return)


; ----------------------------------------------------------------------------

@ReceiveSerial
	(.bcf		STATUS RP0)		; bank 0
	(.btfss		PIR1 RCIF)		; check if data
	(.return)				; if none then return
	(.btfsc		RCSTA OERR)		; if overrun happened
	(.goto		@ErrSerialOver)
	(.btfsc		RCSTA FERR)		; if framing error happend
	(.goto		@ErrSerialFrame)
	(.movf		RCREG W)
	(.return)

@ErrSerialOverr
	(.bcf		RCSTA CREN)		; clear the error
	(.bsf		RCSTA CREN)
	(.return)

@ErrSerialFrame
	(.movf		RCREG W)		; dump it
	(.return)

@TransmitSerial
	(.bcf		STATUS RP0)		; bank 0
	(.btfss		PIR1 TXIF)
	(.goto		(1- @pc)		; Wait until it is done.
	(.movwf		TXREG)
	(.return)


	(.org		#x2100)			; data EEPROM location
	(.dt		"Version 1.0")
	(.de		0)
	(.dt		"Created by: Peter K. Keller (psilord@cs.wisc.edu)")
	(.dt		0)
)
