(defstruct obj
  name
  operand-func
  assembly-func)

(defmacro define-obj (name operands &body body)
  `(defmacro ,name ,operands
     `(make-obj
       :name ',',name
       ;; The arguments to the macro are shoved into a lexically closed thunk
       ;; to be activated later and any unbound variables resolved.
       :operand-func (lambda ()
                       (values
                        ,,@(mapcar (lambda (op)
                                     (if (consp op)
                                         (car op)
                                         op))
                                   (remove '&optional operands))))

       ;; The body of the macro will only be called with the values result
       ;; of the operands-func above. This means only literal values will be
       ;; passed in and the instruction assembled to the byte code.
       :assembly-func (lambda ,',operands
                        ,',@body))))


(restart-case
    (handler-bind ((unbound-variable
                    #'(lambda (c)
                        (format t "Caught condition ~S~%" c)
                        (invoke-restart 'my-restart 7))))
      (format t "Executing body...~%")
      (format t "Value is ~D~%" (+ 1 xxx)))

  (my-restart (&optional v)
    (format t "This is a restart ~D~%" v)))


(defun my-symbol-value (name)
  (if (boundp name)
      (symbol-value name)
      (restart-case (error 'unbound-variable :name name)
        (use-value (value)
          :report "Specify a value to use."
          :interactive (lambda ()
                         (format t "~&Value to use: ")
                         (list (eval (read))))
          value)
        (store-value (value)
          :report "Specify a value to use and store."
          :interactive (lambda ()
                         (format t "~&Value to use and store: ")
                         (list (eval (read))))
          (setf (symbol-value name) value)
          value))))

(handler-bind ((unbound-variable
                #'(lambda (c)      ; argument is condition description
                    ;; Try to make unbound variables get value 17
                    (dolist (tag '(store-value use-value))
                      (let ((restart (find-restart tag c)))
                        (when restart
                          (invoke-restart restart 17)))))))
  (+ (my-symbol-value 'this-symbol-has-no-value)
     (my-symbol-value 'pi)))            ; pi DOES have a value!

(handler-bind ((unbound-variable
                #'(lambda (c)      ; argument is condition description
                    ;; Try to make unbound variables get value 17
                    (dolist (tag '(store-value use-value))
                      (let ((restart (find-restart tag c)))
                        (when restart
                          (invoke-restart restart 17)))))))
  (+ (my-symbol-value 'this-symbol-has-no-value)
     (my-symbol-value 'pi)))            ; pi DOES have a value!

(defun filter-fbound-symbols (expression)
  (mapcan #'(lambda (v)
              (and (not (and (symbolp v) (fboundp v)))
                   (list v)))
          expression))

(defun free-variables (expression &optional (bound-variables '()))
  (cond
    ((symbolp expression)
     (if (member expression bound-variables)
         '()
         (list expression)))

    ((atom expression)
     '())

    ;; Don't wander into a quoted expression!
    ((eq 'quote (car expression))
     '())

    ((eq 'lambda (car expression))
     (let ((bound-variables (append (cadr expression) bound-variables)))
       (mapcan (lambda (subexpr)
                 (free-variables subexpr bound-variables))
               (cddr expression))))

    ((eq 'let (car expression))
     (let ((free-variables
            ;; The free variables in each binding
            (mapcan (lambda (v)
                      (when (not (symbolp v))
                        (free-variables (cdr v) bound-variables)))
                    (cadr expression)))
           ;; The variables bound by this let body.
           (bound-variables
            (append (mapcan (lambda (v)
                              (if (symbolp v)
                                  (list v)
                                  (list (first v))))
                            (cadr expression))
                    ;; Don't forget the previous ones.
                    bound-variables)))

       ;; Process the body with the newly bound symbols in the let
       ;; body.  keep track of all of the free variables from the
       ;; initial bindings plus whatever is in the body of the let.
       (append free-variables
               (mapcan (lambda (subexpr)
                         (free-variables subexpr bound-variables))
                       (cddr expression)))))


    (t
     (mapcan (lambda (subexpr)
               (free-variables subexpr bound-variables))
             expression))))


(defparameter e
  '(let (w (x 0) (y 0))
    (lambda (c) (list w x y z))))


(defmacro flet-dynamic (bindings &body body)
  (let ((names (mapcar #'car bindings))
        (arguments (mapcar #'cadr bindings))
        (bodies (mapcar #'(lambda (b) (cddr b)) bindings))
        (name-syms (loop for i in bindings collect (gensym))))

    ;; bind old values of symbol functions to gensym'ed variables.
    `(let ,(mapcar #'(lambda (ns n) `(,ns (symbol-function ',n)))
                   name-syms names)

       (unwind-protect
            (progn
              ;; reset the symbol functions for named global functions to
              ;; new functions.
              ,@(mapcar #'(lambda (n args bd)
                            `(setf (symbol-function ',n) (lambda ,args ,@bd)))
                        names arguments bodies)

              ;; Execute the body with the new functions.
              ,@body)

         ;; In the cleanup form, reset the global functions to be their
         ;; original settings.
         ,@(mapcar #'(lambda (n ns)
                       `(setf (symbol-function ',n) ,ns))
                   names name-syms)))))


(defun a (x) x)
(defun b () -200)
(defun c (x y) (list x y))

(defun baz-a (x) (a x))
(defun baz-b () (b))
(defun baz-c (x y) (c x y))

(progn
  (format t "[1] (a 10):~D (baz-a 10):~D~%" (a 10) (baz-a 10))
  (format t "[2] (b):~D (baz-b):~D~%" (b) (baz-b))
  (format t "[3] (c 1 2):~A (baz-c 1 2):~A~%" (c 1 2) (baz-c 1 2))

  (flet-dynamic ((a (x) (+ 1 x))
                 (b () 33)
                 (c (x y) (cons x y)))
                (format t "[4] (a 10):~D (baz-a 10):~D~%" (a 10) (baz-a 10))
                (format t "[5] (b):~D (baz-b):~D~%" (b) (baz-b))
                (format t "[6] (c 1 2):~A (baz-c 1 2):~A~%"
                        (c 1 2) (baz-c 1 2)))


  (format t "[7] (a 10):~D (baz-a 10):~D~%" (a 10) (baz-a 10))
  (flet-dynamic ((a (x) (* 10 x)))
                (format t "[8] (a 10):~D (baz-a 10):~D~%" (a 10) (baz-a 10))
                (flet-dynamic ((a (i) (* 100 i)))
                              (format t "[9] (a 10):~D (baz-a 10):~D~%"
                                      (a 10) (baz-a 10))))
  (format t "[10] (a 10):~D (baz-a 10):~D~%" (a 10) (baz-a 10)))




;; read a $symbol and return a form which wraps a restart case around
;; its evaluation.
(eval-when (:compile-toplevel :load-toplevel :execute)
  (set-macro-character
   #\$
   #'(lambda (stream char)
       (declare (ignore char))
       (let ((obj (read stream t nil t)))
         (assert (symbolp obj))
         `(restart-case
              (handler-bind ((warning
                              #'(lambda (c)
                                  (declare (ignore c))
                                  (muffle-warning))))
                (locally #+sbcl(declare (sb-ext:muffle-conditions warning))
                         ,obj))
            (:use-value (value) value))))))

;; Either convert an address into a push of the address object into the
;; instruction stream, or perform a reference to it if we have used it inside
;; of an instruction operand.
(eval-when (:compile-toplevel :load-toplevel :execute)
  (set-macro-character
   #\@
   #'(lambda (stream char)
       (declare (ignore char))
       (let ((obj (read stream t nil t)))
         (assert (symbolp obj))
         `(restart-case
              (handler-bind ((warning
                              #'(lambda (c)
                                  (declare (ignore c))
                                  (muffle-warning))))
                (locally #+sbcl(declare (sb-ext:muffle-conditions warning))
                         ,obj))
            (:use-value (value) value))))))


(defmacro fixup (&body body)
  `(handler-bind ((unbound-variable
                   #'(lambda (c)
                       (declare (special *xxx*))
                       (format t "Fixing up unbound variable: ~A with ~D~%"
                               (cell-error-name c) *xxx*)
                       (invoke-restart :use-value *xxx*))))
     (let ((*xxx* 42))
       (declare (special *xxx*))
       (format t "$foobar is ~D~%" ,@body))))

(defun stuff ()
  (fixup $foobar))

(defun doit ()
  (handler-bind ((unbound-variable
                  #'(lambda (c)
                      (declare (special *xxx*))
                      (format t "Fixing up unbound variable: ~A with ~D~%"
                              (cell-error-name c) *xxx*)
                      (invoke-restart :use-value *xxx*))))
    (with-pic ()
      (let ((*xxx* 42))
        (declare (special *xxx*))

        (.rrf (ash $foobar 1))
        (format t "force operand-func is ~A~%)"
                (funcall (compilable-object-operand-func (car *instruction-stream*))))))))

