(in-package #:pic-instruction-set-12)

(declaim (optimize
          (debug 3)
          (speed 0)
          (compilation-speed 0)))

;; General instruction format for 12-bit wide processors.
(defun byte-oriented (opcode f d)
  (dpb opcode (bits 11 to 6)
       (dpb d (bits 5)
            (ldb (bits 4 to 0) f))))

(defun bit-oriented (opcode f b)
  (dpb opcode (bits 11 to 8)
       (dpb b (bits 7 to 5)
            (ldb (bits 4 to 0) f))))

(defun literal-or-control-operation/no-goto (opcode k)
  (dpb opcode (bits 11 to 8)
       (ldb (bits 7 to 0) k)))

(defun goto-operation (opcode k)
  (dpb opcode (bits 11 to 9)
       (ldb (bits 8 to 0) k)))

(defun gen-label (name)
  (declare (special *current-compilation-unit*))
  (push
   (make-unit
    :kind :label
    :name name)
   (compilation-unit-units *current-compilation-unit*)))

(defmacro .label (name)
  `(gen-label ,name))

(defmacro pic-project (&body body)
  `(let ((*compilation-units* (make-hash-table :test #'equal)))
     (declare (special *compilation-units*))

     ,@body

     *compilation-units*))

(defmacro compilation-unit (style &body body)
  (let ((cun (gensym "COMPILATION-UNIT-"))
        (cu (gensym)))
    `(let ((,cu (make-compilation-unit :style ,style)))
       (declare (special *compilation-units*))
       (setf (gethash ',cun *compilation-units*) ,cu)

       ;; pollute the lexical namespace of the body for macro expansion
       (let ((*current-compilation-unit* ,cu))
         (declare (special *current-compilation-unit*))

         ;; Make sure all @labels which aren't in operands get converted
         ;; into label units and pushed into the current compilation unit.
         (handler-bind ((unbound-variable
                         #'(lambda (c)
			     (format t "Generating Label for ~A~%"
				     (cell-error-name c))
                             (finish-output)
                             (gen-label (cell-error-name c))
                             (invoke-restart :use-value nil))))

           ,@body)

         (with-slots (units environment) *current-compilation-unit*
           (setf units (nreverse units))
           (setf environment (nreverse environment)))

         *current-compilation-unit*))))


(defmacro define-unit (kind name operands &body body)
  `(defmacro ,name ,operands
     `(push
       (make-unit
        :kind ,',kind
        :name ',',name
        ;; The arguments to the macro are shoved into a lexically closed thunk
        ;; to be activated later and any unbound variables resolved.
        :operand-func (lambda ()
                        (values
                         ,,@(mapcar (lambda (op)
                                      (if (consp op)
                                          (car op)
                                          op))
                                    (remove '&optional operands))))

        ;; The body of the macro will only be called with the values result
        ;; of the operands-func above. This means only literal values will be
        ;; passed in and the instruction assembled to the byte code.
        :assembly-func (lambda ,',operands
                         ,',@body))

       (compilation-unit-units *current-compilation-unit*))))

(defmacro define-instruction (name operands &body body)
  `(define-unit :instruction ,name ,operands ,@body))

(defmacro define-directive (name operands &body body)
  `(define-unit :directive ,name ,operands ,@body))


;;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The instruction set for the 12bit opcode series of pic
;;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; All byte oriented instructions
(define-instruction .addwf (f &optional (d #b1))
  (byte-oriented #b000111 f d))

(define-instruction .andwf (f &optional (d #b1))
  (byte-oriented #b000101 f d))

(define-instruction .clrf (f)
  (byte-oriented #b000001 f #b1))

(define-instruction .clrw ()
  (byte-oriented #b000001 #b00000 #b0))

(define-instruction .comf (f &optional (d #b1))
  (byte-oriented #b001001 f d))

(define-instruction .decf (f &optional (d #b1))
  (byte-oriented #b000011 f d))

(define-instruction .decfsz (f &optional (d #b1))
  (byte-oriented #b001011 f d))

(define-instruction .incf (f &optional (d #b1))
  (byte-oriented #b001010 f d))

(define-instruction .incfsz (f &optional (d #b1))
  (byte-oriented #b001111 f d))

(define-instruction .iorwf (f &optional (d #b1))
  (byte-oriented #b000100 f d))

(define-instruction .movf (f &optional (d #b1))
  (byte-oriented #b001000 f d))

(define-instruction .movwf (f)
  (byte-oriented #b000000 f #b1))

(define-instruction .nop ()
  (byte-oriented #b000000 #b00000 #b0))

(define-instruction .rlf (f &optional (d #b1))
  (byte-oriented #b001101 f d))

(define-instruction .rrf (f &optional (d #b1))
  (byte-oriented #b001100 f d))

(define-instruction .subwf (f &optional (d #b1))
  (byte-oriented #b000010 f d))

(define-instruction .swapf (f &optional (d #b1))
  (byte-oriented #b001110 f d))

(define-instruction .xorwf (f &optional (d #b1))
  (byte-oriented #b000110 f d))

;; all bit-oriented instructions
(define-instruction .bcf (f b)
  (bit-oriented #b0100 f b))

(define-instruction .bsf (f b)
  (bit-oriented #b0101 f b))

(define-instruction .btfsc (f b)
  (bit-oriented #b0110 f b))

(define-instruction .btfss (f b)
  (bit-oriented #b0111 f b))

;; Literal and Control operations

(define-instruction .andlw (k)
  (literal-or-control-operation/no-goto #b1110 k))

(define-instruction .call (k)
  (literal-or-control-operation/no-goto #b1001 k))

(define-instruction .clrwdt ()
  (literal-or-control-operation/no-goto #b0000 #b00000100))

(define-instruction .goto (k)
  (goto-operation #b101 k))

(define-instruction .iorlw (k)
  (literal-or-control-operation/no-goto #b1101 k))

(define-instruction .movlw (k)
  (literal-or-control-operation/no-goto #b1100 k))

(define-instruction .option ()
  (literal-or-control-operation/no-goto #b0000 #b00000010))

(define-instruction .retlw (k)
  (literal-or-control-operation/no-goto #b1000 k))

(define-instruction .sleep ()
  (literal-or-control-operation/no-goto #b0000 #b00000011))

;; TRIS is messed up! It is labeled as a literal operation in the
;; manual, but it has a funny syntax.
(define-instruction .tris (f)
  (literal-or-control-operation/no-goto #b0000 (logand f #b111)))

(define-instruction .xorlw (k)
  (literal-or-control-operation/no-goto #b1111 k))



